@@ -7,24 +7,29 @@
  * published by the Free Software Foundation.
  */
 
-#include <stdio.h>
-#include <string.h>
-#include <errno.h>
-#include <time.h>
-#include <ctype.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <sys/socket.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <arpa/inet.h>
-#include <stdarg.h>
-#include <netdb.h>
-#include <dlfcn.h>
-#include <asm/bitops.h>
+#include <stdio.h>			/* *printf, perror, sscanf, fdopen */
+#include <string.h>			/* mem*, str* */
+#include <errno.h>			/* errno, perror */
+#include <time.h>			/* time, ctime */
+#include <netdb.h>			/* gethostby*, getnetby*, getservby* */
+#include <stdlib.h>			/* exit, malloc, free, strtol, getenv, mkstemp */
+#include <unistd.h>			/* read, close, fork, exec*, unlink */
+#include <sys/types.h>			/* open, wait, socket, *sockopt, umask */
+#include <sys/stat.h>			/* open, umask */
+#include <sys/wait.h>			/* wait */
+#include <sys/socket.h>			/* socket, *sockopt, gethostby*, inet_* */
+#include <netinet/in.h>			/* inet_* */
+#include <fcntl.h>			/* open */
+#include <arpa/inet.h>			/* htonl, inet_* */
+#include <stdarg.h>			/* va_* */
+#include <dlfcn.h>			/* dlopen */
 
 #include "ipset.h"
 
+#ifndef PROC_SYS_MODPROBE
+#define PROC_SYS_MODPROBE "/proc/sys/kernel/modprobe"
+#endif
+
 char program_name[] = "ipset";
 char program_version[] = IPSET_VERSION;
 
@@ -42,8 +47,10 @@ static int option_quiet = 0;
 static int restore = 0;
 void *restore_data = NULL;
 struct ip_set_restore *restore_set = NULL;
-size_t restore_offset = 0, restore_size;
-unsigned line = 0;
+size_t restore_offset = 0;
+socklen_t restore_size;
+unsigned restore_line = 0;
+unsigned warn_once = 0;
 
 #define TEMPFILE_PATTERN	"/ipsetXXXXXX"
 
@@ -68,9 +75,10 @@ static const char cmdflags[] = { ' ',			/* CMD_NONE */
 #define OPT_QUIET		0x0004U		/* -q */
 #define OPT_DEBUG		0x0008U		/* -z */
 #define OPT_BINDING		0x0010U		/* -b */
-#define NUMBER_OF_OPT 5
+#define OPT_RESOLVE		0x0020U		/* -r */
+#define NUMBER_OF_OPT 6
 static const char optflags[] =
-    { 'n', 's', 'q', 'z', 'b' };
+    { 'n', 's', 'q', 'z', 'b', 'r' };
 
 static struct option opts_long[] = {
 	/* set operations */
@@ -98,6 +106,7 @@ static struct option opts_long[] = {
 	{"sorted",  0, 0, 's'},
 	{"quiet",   0, 0, 'q'},
 	{"binding", 1, 0, 'b'},
+	{"resolve", 0, 0, 'r'},
 
 #ifdef IPSET_DEBUG
 	/* debug (if compiled with it) */
@@ -109,11 +118,11 @@ static struct option opts_long[] = {
 	{"help",    2, 0, 'H'},
 
 	/* end */
-	{0}
+	{NULL},
 };
 
 static char opts_short[] =
-    "-N:X::F::E:W:L::S::RA:D:T:B:U:nsqzb:Vh::H::";
+    "-N:X::F::E:W:L::S::RA:D:T:B:U:nrsqzb:Vh::H::";
 
 /* Table of legal combinations of commands and options. If any of the
  * given commands make an option legal, that option is legal.
@@ -125,27 +134,27 @@ static char opts_short[] =
 
 static char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] = {
 	/*            -n   -s   -q   -z   -b  */
-	 /*CREATE*/  {'x', 'x', ' ', ' ', 'x'},
-	 /*DESTROY*/ {'x', 'x', ' ', ' ', 'x'},
-	 /*FLUSH*/   {'x', 'x', ' ', ' ', 'x'},
-	 /*RENAME*/  {'x', 'x', ' ', ' ', 'x'},
-	 /*SWAP*/    {'x', 'x', ' ', ' ', 'x'},
-	 /*LIST*/    {' ', ' ', 'x', ' ', 'x'},
-	 /*SAVE*/    {'x', 'x', ' ', ' ', 'x'},
-	 /*RESTORE*/ {'x', 'x', ' ', ' ', 'x'},
-	 /*ADD*/     {'x', 'x', ' ', ' ', 'x'},
-	 /*DEL*/     {'x', 'x', ' ', ' ', 'x'},
-	 /*TEST*/    {'x', 'x', ' ', ' ', ' '},
-	 /*BIND*/    {'x', 'x', ' ', ' ', '+'},
-	 /*UNBIND*/  {'x', 'x', ' ', ' ', 'x'},
-	 /*HELP*/    {'x', 'x', 'x', ' ', 'x'},
-	 /*VERSION*/ {'x', 'x', 'x', ' ', 'x'},
+	 /*CREATE*/  {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*DESTROY*/ {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*FLUSH*/   {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*RENAME*/  {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*SWAP*/    {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*LIST*/    {' ', ' ', 'x', ' ', 'x', ' '},
+	 /*SAVE*/    {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*RESTORE*/ {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*ADD*/     {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*DEL*/     {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*TEST*/    {'x', 'x', ' ', ' ', ' ', 'x'},
+	 /*BIND*/    {'x', 'x', ' ', ' ', '+', 'x'},
+	 /*UNBIND*/  {'x', 'x', ' ', ' ', 'x', 'x'},
+	 /*HELP*/    {'x', 'x', 'x', ' ', 'x', 'x'},
+	 /*VERSION*/ {'x', 'x', 'x', ' ', 'x', 'x'},
 };
 
 /* Main parser function */
 int parse_commandline(int argc, char *argv[]);
 
-void exit_tryhelp(int status)
+static void exit_tryhelp(int status)
 {
 	fprintf(stderr,
 		"Try `%s -H' or '%s --help' for more information.\n",
@@ -153,18 +162,18 @@ void exit_tryhelp(int status)
 	exit(status);
 }
 
-void exit_error(enum exittype status, char *msg, ...)
+void exit_error(int status, const char *msg, ...)
 {
-	va_list args;
-
 	if (!option_quiet) {
+		va_list args;
+
 		va_start(args, msg);
 		fprintf(stderr, "%s v%s: ", program_name, program_version);
 		vfprintf(stderr, msg, args);
 		va_end(args);
 		fprintf(stderr, "\n");
-		if (line)
-			fprintf(stderr, "Restore failed at line %u:\n", line);
+		if (restore_line)
+			fprintf(stderr, "Restore failed at line %u:\n", restore_line);
 		if (status == PARAMETER_PROBLEM)
 			exit_tryhelp(status);
 		if (status == VERSION_PROBLEM)
@@ -176,11 +185,11 @@ void exit_error(enum exittype status, char *msg, ...)
 	exit(status);
 }
 
-void ipset_printf(char *msg, ...)
+static void ipset_printf(const char *msg, ...)
 {
-	va_list args;
-
 	if (!option_quiet) {
+		va_list args;
+
 		va_start(args, msg);
 		vfprintf(stdout, msg, args);
 		va_end(args);
@@ -188,7 +197,7 @@ void ipset_printf(char *msg, ...)
 	}
 }
 
-static void generic_opt_check(int command, int options)
+static void generic_opt_check(int command, unsigned int options)
 {
 	int i, j, legal = 0;
 
@@ -223,7 +232,7 @@ static void generic_opt_check(int command, int options)
 	}
 }
 
-static char opt2char(int option)
+static char opt2char(unsigned int option)
 {
 	const char *ptr;
 	for (ptr = optflags; option > 1; option >>= 1, ptr++);
@@ -231,12 +240,80 @@ static char opt2char(int option)
 	return *ptr;
 }
 
-static char cmd2char(int option)
+static char cmd2char(int cmd)
 {
-	if (option <= CMD_NONE || option > NUMBER_OF_CMD)
+	if (cmd <= CMD_NONE || cmd > NUMBER_OF_CMD)
 		return ' '; 
 
-	return cmdflags[option];
+	return cmdflags[cmd];
+}
+
+/* From iptables.c ... */
+static char *get_modprobe(void)
+{
+	int procfile;
+	char *ret;
+
+#define PROCFILE_BUFSIZ	1024
+	procfile = open(PROC_SYS_MODPROBE, O_RDONLY);
+	if (procfile < 0)
+		return NULL;
+
+	ret = (char *) malloc(PROCFILE_BUFSIZ);
+	if (ret) {
+		memset(ret, 0, PROCFILE_BUFSIZ);
+		switch (read(procfile, ret, PROCFILE_BUFSIZ)) {
+		case -1: goto fail;
+		case PROCFILE_BUFSIZ: goto fail; /* Partial read.  Wierd */
+		default: ; /* nothing */
+		}
+		if (ret[strlen(ret)-1]=='\n') 
+			ret[strlen(ret)-1]=0;
+		close(procfile);
+		return ret;
+	}
+ fail:
+	free(ret);
+	close(procfile);
+	return NULL;
+}
+
+static int ipset_insmod(const char *modname, const char *modprobe)
+{
+	char *buf = NULL;
+	char *argv[3];
+	struct stat junk;
+	int status;
+	
+	if (!stat(modprobe, &junk)) {
+		/* Try to read out of the kernel */
+		buf = get_modprobe();
+		if (!buf)
+			return -1;
+		modprobe = buf;
+	}
+	
+	switch (fork()) {
+	case 0:
+		argv[0] = (char *) modprobe;
+		argv[1] = (char *) modname;
+		argv[2] = NULL;
+		execv(argv[0], argv);
+		
+		/* Should not reach */
+		exit(1);
+	case -1:
+		return -1;
+	
+	default: /* parent */
+		wait(&status);
+	}
+	
+	free(buf);
+	
+	if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
+		return 0;
+	return -1;
 }
 
 static int kernel_getsocket(void)
@@ -257,7 +334,7 @@ static void kernel_error(unsigned cmd, int err)
 	struct translate_error {
 		int err;
 		unsigned cmd;
-		char *message;
+		const char *message;
 	} table[] =
 	{ /* Generic error codes */
 	  { EPERM, 0, "Missing capability" },
@@ -275,11 +352,12 @@ static void kernel_error(unsigned cmd, int err)
 	  { ENOENT, 0, "Unknown set" },
 	  { EAGAIN, 0, "Sets are busy, try again later" },
 	  { ERANGE, CMD_CREATE, "No free slot remained to add a new set" },
-	  { ERANGE, 0, "IP/port is outside of the set" },
+	  { ERANGE, 0, "IP/port/element is outside of the set or set is full" },
 	  { ENOEXEC, CMD_CREATE, "Invalid parameters to create a set" },
 	  { ENOEXEC, CMD_SWAP, "Sets with different types cannot be swapped" },
 	  { EEXIST, CMD_CREATE, "Set already exists" },
 	  { EEXIST, CMD_RENAME, "Set with new name already exists" },
+	  { EEXIST, 0, "Set specified as element does not exist" },
 	  { EBUSY, 0, "Set is in use, operation not permitted" },
 	  };
 	for (i = 0; i < sizeof(table)/sizeof(struct translate_error); i++) {
@@ -292,30 +370,50 @@ static void kernel_error(unsigned cmd, int err)
 	exit_error(OTHER_PROBLEM, "Error from kernel: %s", strerror(err));
 }
 
-static void kernel_getfrom(unsigned cmd, void *data, size_t * size)
+static inline int wrapped_getsockopt(void *data, socklen_t *size)
 {
 	int res;
 	int sockfd = kernel_getsocket();
 
 	/* Send! */
 	res = getsockopt(sockfd, SOL_IP, SO_IP_SET, data, size);
-
+	if (res != 0 
+	    && errno == ENOPROTOOPT 
+	    && ipset_insmod("ip_set", "/sbin/modprobe") == 0)
+		res = getsockopt(sockfd, SOL_IP, SO_IP_SET, data, size);
 	DP("res=%d errno=%d", res, errno);
-
-	if (res != 0)
-		kernel_error(cmd, errno);
+	
+	return res;
 }
 
-static int kernel_sendto_handleerrno(unsigned cmd, unsigned op,
-				     void *data, size_t size)
+static inline int wrapped_setsockopt(void *data, socklen_t size)
 {
 	int res;
 	int sockfd = kernel_getsocket();
 
 	/* Send! */
 	res = setsockopt(sockfd, SOL_IP, SO_IP_SET, data, size);
-
+	if (res != 0 
+	    && errno == ENOPROTOOPT 
+	    && ipset_insmod("ip_set", "/sbin/modprobe") == 0)
+		res = setsockopt(sockfd, SOL_IP, SO_IP_SET, data, size);
 	DP("res=%d errno=%d", res, errno);
+	
+	return res;
+}
+
+static void kernel_getfrom(unsigned cmd, void *data, socklen_t * size)
+{
+	int res = wrapped_getsockopt(data, size);
+
+	if (res != 0)
+		kernel_error(cmd, errno);
+}
+
+static int kernel_sendto_handleerrno(unsigned cmd,
+				     void *data, socklen_t size)
+{
+	int res = wrapped_setsockopt(data, size);
 
 	if (res != 0) {
 		if (errno == EEXIST)
@@ -329,27 +427,15 @@ static int kernel_sendto_handleerrno(unsigned cmd, unsigned op,
 
 static void kernel_sendto(unsigned cmd, void *data, size_t size)
 {
-	int res;
-	int sockfd = kernel_getsocket();
-
-	/* Send! */
-	res = setsockopt(sockfd, SOL_IP, SO_IP_SET, data, size);
-
-	DP("res=%d errno=%d", res, errno);
+	int res = wrapped_setsockopt(data, size);
 
 	if (res != 0)
 		kernel_error(cmd, errno);
 }
 
-static int kernel_getfrom_handleerrno(unsigned cmd, void *data, size_t * size)
+static int kernel_getfrom_handleerrno(unsigned cmd, void *data, socklen_t *size)
 {
-	int res;
-	int sockfd = kernel_getsocket();
-
-	/* Send! */
-	res = getsockopt(sockfd, SOL_IP, SO_IP_SET, data, size);
-
-	DP("res=%d errno=%d", res, errno);
+	int res = wrapped_getsockopt(data, size);
 
 	if (res != 0) {
 		if (errno == EAGAIN)
@@ -364,7 +450,7 @@ static int kernel_getfrom_handleerrno(unsigned cmd, void *data, size_t * size)
 static void check_protocolversion(void)
 {
 	struct ip_set_req_version req_version;
-	size_t size = sizeof(struct ip_set_req_version);
+	socklen_t size = sizeof(struct ip_set_req_version);
 	int sockfd = kernel_getsocket();
 	int res;
 
@@ -386,7 +472,7 @@ static void check_protocolversion(void)
 			   req_version.version, IP_SET_PROTOCOL_VERSION);
 }
 
-static void set_command(int *cmd, const int newcmd)
+static void set_command(int *cmd, int newcmd)
 {
 	if (*cmd != CMD_NONE)
 		exit_error(PARAMETER_PROBLEM, "Can't use -%c with -%c\n",
@@ -417,18 +503,28 @@ void *ipset_malloc(size_t size)
 	return p;
 }
 
-void ipset_free(void **data)
+char *ipset_strdup(const char *s)
 {
-	if (*data == NULL)
+	char *p;
+
+	if ((p = strdup(s)) == NULL) {
+		perror("ipset: not enough memory");
+		exit(1);
+	}
+	return p;
+}
+
+void ipset_free(void *data)
+{
+	if (data == NULL)
 		return;
 
-	free(*data);
-	*data = NULL;
+	free(data);
 }
 
 static struct option *merge_options(struct option *oldopts,
 				    const struct option *newopts,
-				    unsigned int *option_offset)
+				    int *option_offset)
 {
 	unsigned int num_old, num_new, i;
 	struct option *merge;
@@ -439,7 +535,7 @@ static struct option *merge_options(struct option *oldopts,
 	global_option_offset += OPTION_OFFSET;
 	*option_offset = global_option_offset;
 
-	merge = malloc(sizeof(struct option) * (num_new + num_old + 1));
+	merge = ipset_malloc(sizeof(struct option) * (num_new + num_old + 1));
 	memcpy(merge, oldopts, num_old * sizeof(struct option));
 	for (i = 0; i < num_new; i++) {
 		merge[num_old + i] = newopts[i];
@@ -468,7 +564,7 @@ static char *ip_tonetwork(const struct in_addr *addr)
 {
 	struct netent *net;
 
-	if ((net = getnetbyaddr((long) ntohl(addr->s_addr), 
+	if ((net = getnetbyaddr(ntohl(addr->s_addr), 
 				AF_INET)) != NULL) {
 		DP("%s", net->n_name);
 		return (char *) net->n_name;
@@ -483,19 +579,23 @@ static char *ip_tonetwork(const struct in_addr *addr)
 char *ip_tostring(ip_set_ip_t ip, unsigned options)
 {
 	struct in_addr addr;
-	char *name;
-
 	addr.s_addr = htonl(ip);
 
 	if (!(options & OPT_NUMERIC)) {
+		char *name;
 		if ((name = ip_tohost(&addr)) != NULL ||
 		    (name = ip_tonetwork(&addr)) != NULL)
 			return name;
 	}
-
+	
 	return inet_ntoa(addr);
 }
 
+char *binding_ip_tostring(struct set *set UNUSED,
+			  ip_set_ip_t ip, unsigned options)
+{
+	return ip_tostring(ip, options);
+}
 char *ip_tostring_numeric(ip_set_ip_t ip)
 {
 	return ip_tostring(ip, OPT_NUMERIC);
@@ -511,9 +611,6 @@ void parse_ip(const char *str, ip_set_ip_t * ip)
 	
 	if (inet_aton(str, &addr) != 0) {
 		*ip = ntohl(addr.s_addr);	/* We want host byte order */
-		if (!*ip)
-			exit_error(PARAMETER_PROBLEM,
-				   "Zero valued IP address `%s' specified", str);
 		return;
 	}
 
@@ -530,10 +627,6 @@ void parse_ip(const char *str, ip_set_ip_t * ip)
 				   "Please specify one.", str);
 
 		*ip = ntohl(((struct in_addr *) host->h_addr_list[0])->s_addr);
-		if (!*ip)
-			exit_error(PARAMETER_PROBLEM,
-				   "Zero valued IP address `%s' specified",
-				   str);
 		return;
 	}
 
@@ -544,7 +637,7 @@ void parse_ip(const char *str, ip_set_ip_t * ip)
 void parse_mask(const char *str, ip_set_ip_t * mask)
 {
 	struct in_addr addr;
-	unsigned int bits;
+	int bits;
 
 	DP("%s", str);
 
@@ -563,7 +656,7 @@ void parse_mask(const char *str, ip_set_ip_t * mask)
 
 	DP("bits: %d", bits);
 
-	*mask = 0xFFFFFFFF << (32 - bits);
+	*mask = bits != 0 ? 0xFFFFFFFF << (32 - bits) : 0L;
 }
 
 /* Combines parse_ip and parse_mask */
@@ -589,13 +682,13 @@ parse_ipandmask(const char *str, ip_set_ip_t * ip, ip_set_ip_t * mask)
 		parse_ip(buf, ip);
 
 	DP("%s ip: %08X (%s) mask: %08X",
-	   str, *ip, ip_tostring(*ip, 0), *mask);
+	   str, *ip, ip_tostring_numeric(*ip), *mask);
 
 	/* Apply the netmask */
 	*ip &= *mask;
 
 	DP("%s ip: %08X (%s) mask: %08X",
-	   str, *ip, ip_tostring(*ip, 0), *mask);
+	   str, *ip, ip_tostring_numeric(*ip), *mask);
 }
 
 /* Return a string representation of a port
@@ -619,12 +712,12 @@ int
 string_to_number(const char *str, unsigned int min, unsigned int max,
 		 ip_set_ip_t *port)
 {
-	long number;
+	unsigned long number;
 	char *end;
 
 	/* Handle hex, octal, etc. */
 	errno = 0;
-	number = strtol(str, &end, 0);
+	number = strtoul(str, &end, 0);
 	if (*end == '\0' && end != str) {
 		/* we parsed a number, let's see if we want this */
 		if (errno != ERANGE && min <= number && number <= max) {
@@ -639,12 +732,11 @@ static int
 string_to_port(const char *str, ip_set_ip_t *port)
 {
 	struct servent *service;
-
+	
 	if ((service = getservbyname(str, "tcp")) != NULL) {
-		*port = ntohs((unsigned short) service->s_port);
+		*port = ntohs((uint16_t) service->s_port);
 		return 0;
 	}
-	
 	return -1;
 }
 
@@ -654,11 +746,7 @@ void parse_port(const char *str, ip_set_ip_t *port)
 	if ((string_to_number(str, 0, 65535, port) != 0)
 	      && (string_to_port(str, port) != 0))
 		exit_error(PARAMETER_PROBLEM, 
-		           "Invalid TCP port `%s' specified", str);
-	
-	if (!*port)
-		exit_error(PARAMETER_PROBLEM, 
-		           "Zero valued port `%s' specified", str);
+		           "Invalid TCP port `%s' specified", str);	
 }
 
 /* 
@@ -771,7 +859,7 @@ void settype_register(struct settype *settype)
 }
 
 /* Find set functions */
-static struct set *set_find_byid(ip_set_id_t id)
+struct set *set_find_byid(ip_set_id_t id)
 {
 	struct set *set = NULL;
 	ip_set_id_t i;
@@ -788,7 +876,7 @@ static struct set *set_find_byid(ip_set_id_t id)
 	return set;
 }
 
-static struct set *set_find_byname(const char *name)
+struct set *set_find_byname(const char *name)
 {
 	struct set *set = NULL;
 	ip_set_id_t i;
@@ -808,12 +896,12 @@ static struct set *set_find_byname(const char *name)
 
 static ip_set_id_t set_find_free_index(const char *name)
 {
-	ip_set_id_t i, index = IP_SET_INVALID_ID;
+	ip_set_id_t i, idx = IP_SET_INVALID_ID;
 
 	for (i = 0; i < max_sets; i++) {
-		if (index == IP_SET_INVALID_ID
+		if (idx == IP_SET_INVALID_ID
 		    && set_list[i] == NULL)
-			index = i;
+			idx = i;
 		if (set_list[i] != NULL
 		    && strncmp(set_list[i]->name, name,
 			       IP_SET_MAXNAMELEN) == 0)
@@ -822,13 +910,13 @@ static ip_set_id_t set_find_free_index(const char *name)
    				   name);
 	}
 			
-	if (index == IP_SET_INVALID_ID)		
+	if (idx == IP_SET_INVALID_ID)		
 		exit_error(PARAMETER_PROBLEM,
 	   		   "Set %s cannot be restored, "
 	   		   "max number of set %u reached",
 	   		   name, max_sets);
 
-	return index;
+	return idx;
 }
 
 /* 
@@ -947,7 +1035,7 @@ static void set_rename(const char *name, const char *newname,
  * Send MAX_SETS, LIST_SIZE and/or SAVE_SIZE orders to kernel
  */
 static size_t load_set_list(const char name[IP_SET_MAXNAMELEN],
-			    ip_set_id_t *index,
+			    ip_set_id_t *idx,
 			    unsigned op, unsigned cmd)
 {
 	void *data = NULL;
@@ -955,7 +1043,7 @@ static size_t load_set_list(const char name[IP_SET_MAXNAMELEN],
 	struct ip_set_name_list *name_list;
 	struct set *set;
 	ip_set_id_t i;
-	size_t size, req_size;
+	socklen_t size, req_size;
 	int repeated = 0, res = 0;
 
 	DP("%s %s", cmd == CMD_MAX_SETS ? "MAX_SETS"
@@ -984,7 +1072,7 @@ tryagain:
 	max_sets = req_max_sets.max_sets;
 	set_list = ipset_malloc(max_sets * sizeof(struct set *));
 	memset(set_list, 0, max_sets * sizeof(struct set *));
-	*index = req_max_sets.set.index;
+	*idx = req_max_sets.set.index;
 
 	if (req_max_sets.sets == 0)
 		/* No sets in kernel */
@@ -995,7 +1083,7 @@ tryagain:
 			  + req_max_sets.sets * sizeof(struct ip_set_name_list);
 	data = ipset_malloc(size);
 	((struct ip_set_req_setnames *) data)->op = op;
-	((struct ip_set_req_setnames *) data)->index = *index;
+	((struct ip_set_req_setnames *) data)->index = *idx;
 
 	res = kernel_getfrom_handleerrno(cmd, data, &size);
 
@@ -1049,9 +1137,14 @@ static size_t save_bindings(void *data, size_t offset, size_t len)
 	if (!(set && set_list[hash->binding]))
 		exit_error(OTHER_PROBLEM,
 			   "Save binding failed, try again later.");
+	if (!set->settype->bindip_tostring)
+		exit_error(OTHER_PROBLEM,
+			   "Internal error, binding is not supported with set %s"
+			   " of settype %s\n",
+			   set->name, set->settype->typename);
 	printf("-B %s %s -b %s\n",
 		set->name,
-		set->settype->bindip_tostring(hash->ip, OPT_NUMERIC),
+		set->settype->bindip_tostring(set, hash->ip, OPT_NUMERIC),
 		set_list[hash->binding]->name);
 
 	return sizeof(struct ip_set_hash_save);
@@ -1066,13 +1159,17 @@ static size_t save_set(void *data, int *bindings,
 	struct settype *settype;
 	size_t used;
 	
-	DP("offset %u, len %u", offset, len);
+	DP("offset %u (%u/%u/%u), len %u", offset,
+	   sizeof(struct ip_set_save), 
+	   set_save->header_size, set_save->members_size, 
+	   len);
 	if (offset + sizeof(struct ip_set_save) > len
 	    || offset + sizeof(struct ip_set_save)
 	       + set_save->header_size + set_save->members_size > len)
 		exit_error(OTHER_PROBLEM,
 			   "Save operation failed, try again later.");
 
+	DP("index: %u", set_save->index);
 	if (set_save->index == IP_SET_INVALID_ID) {
 		/* Marker */
 		*bindings = 1;
@@ -1129,13 +1226,13 @@ static size_t save_default_bindings(void *data, int *bindings)
 static int try_save_sets(const char name[IP_SET_MAXNAMELEN])
 {
 	void *data = NULL;
-	size_t size, req_size = 0;
-	ip_set_id_t index;
+	socklen_t size, req_size = 0;
+	ip_set_id_t idx;
 	int res = 0, bindings = 0;
 	time_t now = time(NULL);
 
 	/* Load set_list from kernel */
-	size = load_set_list(name, &index,
+	size = load_set_list(name, &idx,
 			     IP_SET_OP_SAVE_SIZE, CMD_SAVE);
 	
 	if (size) {
@@ -1144,7 +1241,7 @@ static int try_save_sets(const char name[IP_SET_MAXNAMELEN])
 		req_size = (size += sizeof(struct ip_set_save));
 		data = ipset_malloc(size);
 		((struct ip_set_req_list *) data)->op = IP_SET_OP_SAVE;
-		((struct ip_set_req_list *) data)->index = index;
+		((struct ip_set_req_list *) data)->index = idx;
 		res = kernel_getfrom_handleerrno(CMD_SAVE, data, &size);
 
 		if (res != 0 || size != req_size) {
@@ -1171,7 +1268,7 @@ static int try_save_sets(const char name[IP_SET_MAXNAMELEN])
 	printf("COMMIT\n");
 	now = time(NULL);
 	printf("# Completed on %s", ctime(&now));
-	ipset_free(&data);
+	ipset_free(data);
 	return res;
 }
 
@@ -1205,7 +1302,7 @@ static char *newargv[255];
 static int newargc = 0;
 
 /* Build faked argv from parsed line */
-static void build_argv(int line, char *buffer) {
+static void build_argv(unsigned line, char *buffer) {
 	char *ptr;
 	int i;
 
@@ -1215,10 +1312,10 @@ static void build_argv(int line, char *buffer) {
 	newargc = 1;
 
 	ptr = strtok(buffer, " \t\n");
-	newargv[newargc++] = strdup(ptr);
+	newargv[newargc++] = ipset_strdup(ptr);
 	while ((ptr = strtok(NULL, " \t\n")) != NULL) {
-		if ((newargc + 1) < sizeof(newargv)/sizeof(char *))
-			newargv[newargc++] = strdup(ptr);
+		if ((newargc + 1) < (int)(sizeof(newargv)/sizeof(char *)))
+			newargv[newargc++] = ipset_strdup(ptr);
 		else
 			exit_error(PARAMETER_PROBLEM,
 				   "Line %d is too long to restore\n", line);
@@ -1227,19 +1324,17 @@ static void build_argv(int line, char *buffer) {
 
 static FILE *create_tempfile(void)
 {
-	char buffer[1024];	
+	char buffer[1024], __tmpdir[] = "/tmp";	
 	char *tmpdir = NULL;
 	char *filename;
 	int fd;
 	FILE *file;
 	
 	if (!(tmpdir = getenv("TMPDIR")) && !(tmpdir = getenv("TMP")))
-		tmpdir = "/tmp";
-	filename = malloc(strlen(tmpdir) + strlen(TEMPFILE_PATTERN) + 1);
-	if (!filename)
-		exit_error(OTHER_PROBLEM, "Could not malloc temporary filename.");
+		tmpdir = __tmpdir;
+	filename = ipset_malloc(strlen(tmpdir) + strlen(TEMPFILE_PATTERN) + 1);
 	strcpy(filename, tmpdir);
-	strcpy(filename, TEMPFILE_PATTERN);
+	strcat(filename, TEMPFILE_PATTERN);
 	
 	(void) umask(077);	/* Create with restrictive permissions */
 	fd = mkstemp(filename);
@@ -1267,10 +1362,10 @@ static void set_restore(char *argv0)
 	char buffer[1024];	
 	char *ptr, *name = NULL;
 	char cmd = ' ';
-	int line = 0, first_pass, i, bindings = 0;
+	int first_pass, i, bindings = 0;
 	struct settype *settype = NULL;
 	struct ip_set_req_setnames *header;
-	ip_set_id_t index;
+	ip_set_id_t idx;
 	FILE *in;
 	int res;
 	
@@ -1278,15 +1373,16 @@ static void set_restore(char *argv0)
 	in = create_tempfile();
 	
 	/* Load existing sets from kernel */
-	load_set_list(IPSET_TOKEN_ALL, &index,
+	load_set_list(IPSET_TOKEN_ALL, &idx,
 		      IP_SET_OP_LIST_SIZE, CMD_RESTORE);
 	
+	restore_line = 0;
 	restore_size = sizeof(struct ip_set_req_setnames)/* header */
 		       + sizeof(struct ip_set_restore);  /* marker */
 	DP("restore_size: %u", restore_size);
 	/* First pass: calculate required amount of data */
 	while (fgets(buffer, sizeof(buffer), in)) {
-		line++;
+		restore_line++;
 
 		if (buffer[0] == '\n')
 			continue;
@@ -1309,7 +1405,7 @@ static void set_restore(char *argv0)
 		    || ptr[2] != '\0') {
 			exit_error(PARAMETER_PROBLEM,
 				   "Line %u does not start as a valid restore command\n",
-				   line);
+				   restore_line);
 		}
 		cmd = ptr[1];		
 		/* setname */
@@ -1318,7 +1414,7 @@ static void set_restore(char *argv0)
 		if (ptr == NULL)
 		        exit_error(PARAMETER_PROBLEM,
 		        	   "Missing set name in line %u\n",
-		        	   line);
+		        	   restore_line);
 		DP("cmd %c", cmd);
 		switch (cmd) {
 		case 'N': {
@@ -1328,11 +1424,11 @@ static void set_restore(char *argv0)
 			if (ptr == NULL)
 			        exit_error(PARAMETER_PROBLEM,
 			        	   "Missing settype in line %u\n",
-		        		   line);
+		        		   restore_line);
 			if (bindings)
 			        exit_error(PARAMETER_PROBLEM,
 			        	   "Invalid line %u: create must precede bindings\n",
-		        		   line);
+		        		   restore_line);
 			settype = check_set_typename(ptr);
 			restore_size += sizeof(struct ip_set_restore)
 					+ settype->create_size;
@@ -1345,11 +1441,11 @@ static void set_restore(char *argv0)
 			        exit_error(PARAMETER_PROBLEM,
 			        	   "Add IP to set %s in line %u without "
 					   "preceding corresponding create set line\n",
-		        		   ptr, line);
+		        		   ptr, restore_line);
 			if (bindings)
 			        exit_error(PARAMETER_PROBLEM,
 			        	   "Invalid line %u: adding entries must precede bindings\n",
-		        		   line);
+		        		   restore_line);
 			restore_size += settype->adt_size;
 			DP("restore_size (A): %u", restore_size);
 			break;
@@ -1363,7 +1459,7 @@ static void set_restore(char *argv0)
 		default: {
 			exit_error(PARAMETER_PROBLEM,
 		       		   "Unrecognized restore command in line %u\n",
-				   line);
+				   restore_line);
 		}
 		} /* end of switch */
 	}			
@@ -1380,15 +1476,15 @@ static void set_restore(char *argv0)
 
 	/* Rewind to scan the file again */
 	fseek(in, 0L, SEEK_SET);
-	first_pass = line;
-	line = 0;
+	first_pass = restore_line;
+	restore_line = 0;
 	
 	/* Initialize newargv/newargc */
-	newargv[newargc++] = strdup(argv0);
+	newargv[newargc++] = ipset_strdup(argv0);
 	
 	/* Second pass: build up restore request */
 	while (fgets(buffer, sizeof(buffer), in)) {		
-		line++;
+		restore_line++;
 
 		if (buffer[0] == '\n')
 			continue;
@@ -1398,7 +1494,7 @@ static void set_restore(char *argv0)
 			goto do_restore;
 		DP("restoring: %s", buffer);
 		/* Build faked argv, argc */
-		build_argv(line, buffer);
+		build_argv(restore_line, buffer);
 		for (i = 0; i < newargc; i++)
 			DP("argv[%u]: %s", i, newargv[i]);
 		
@@ -1446,7 +1542,7 @@ static struct set *set_adt_get(const char *name)
 {
 	struct ip_set_req_adt_get req_adt_get;
 	struct set *set;
-	size_t size;
+	socklen_t size;
 
 	DP("%s", name);
 
@@ -1455,7 +1551,7 @@ static struct set *set_adt_get(const char *name)
 	strcpy(req_adt_get.set.name, name);
 	size = sizeof(struct ip_set_req_adt_get);
 
-	kernel_getfrom(CMD_ADT_GET, &req_adt_get, &size);
+	kernel_getfrom(CMD_ADT_GET, (void *) &req_adt_get, &size);
 
 	set = ipset_malloc(sizeof(struct set));
 	strcpy(set->name, name);
@@ -1490,7 +1586,7 @@ static int set_adtip(struct set *set, const char *adt,
 	memcpy(data + sizeof(struct ip_set_req_adt),
 	       set->settype->data, set->settype->adt_size);
 	
-	if (kernel_sendto_handleerrno(cmd, op, data, size) == -1)
+	if (kernel_sendto_handleerrno(cmd, data, size) == -1)
 		switch (op) {
 		case IP_SET_OP_ADD_IP:
 			exit_error(OTHER_PROBLEM, "%s is already in set %s.",
@@ -1521,7 +1617,7 @@ static int set_adtip(struct set *set, const char *adt,
 	return res;
 }
 
-static void set_restore_add(struct set *set, const char *adt)
+static void set_restore_add(struct set *set, const char *adt UNUSED)
 {
 	DP("%s %s", set->name, adt);
 	/* Sanity checking */
@@ -1550,6 +1646,15 @@ static int set_bind(struct set *set, const char *adt,
 	/* set may be null: '-U :all: :all:|:default:' */
 	DP("(%s, %s) -> %s", set ? set->name : IPSET_TOKEN_ALL, adt, binding);
 
+	/* Ugly */
+	if (set != NULL
+	    && ((strcmp(set->settype->typename, "iptreemap") == 0)
+	        || (strcmp(set->settype->typename, "ipportiphash") == 0)
+	        || (strcmp(set->settype->typename, "ipportnethash") == 0)
+	        || (strcmp(set->settype->typename, "setlist") == 0)))
+		exit_error(PARAMETER_PROBLEM,
+			"%s type of sets cannot be used at binding operations\n",
+			set->settype->typename);
 	/* Alloc memory for the data to send */
 	size = sizeof(struct ip_set_req_bind);
 	if (op != IP_SET_OP_UNBIND_SET && adt[0] == ':')
@@ -1577,7 +1682,7 @@ static int set_bind(struct set *set, const char *adt,
 	}
 
 	if (op == IP_SET_OP_TEST_BIND_SET) {
-		if (kernel_sendto_handleerrno(cmd, op, data, size) == -1) {
+		if (kernel_sendto_handleerrno(cmd, data, size) == -1) {
 			ipset_printf("%s in set %s is bound to %s.",
 				     adt, set->name, binding);
 			res = 0;
@@ -1623,8 +1728,14 @@ static void set_restore_bind(struct set *set,
 	DP("%s -> %s", adt, binding);
 	if (strcmp(adt, IPSET_TOKEN_DEFAULT) == 0)
 		hash_restore->ip = 0;
-	else
+	else {
+		if (!set->settype->bindip_parse)
+			exit_error(OTHER_PROBLEM,
+				   "Internal error, binding is not supported with set %s"
+				   " of settype %s\n",
+				   set->name, set->settype->typename);
 		set->settype->bindip_parse(adt, &hash_restore->ip);
+	}
 	hash_restore->id = set->index;	    		   
 	hash_restore->binding = (set_find_byname(binding))->index;	
 	DP("id %u, ip %u, binding %u",
@@ -1636,16 +1747,24 @@ static void set_restore_bind(struct set *set,
  * Print operation
  */
 
-static void print_bindings(void *data, size_t size, unsigned options,
-			   char * (*printip)(ip_set_ip_t ip, unsigned options))
+static void print_bindings(struct set *set,
+			   void *data, size_t size, unsigned options,
+			   char * (*printip)(struct set *set, 
+					     ip_set_ip_t ip, unsigned options))
 {
 	size_t offset = 0;
 	struct ip_set_hash_list *hash;
 
+	if (offset < size && !printip)
+		exit_error(OTHER_PROBLEM,
+			   "Internal error, binding is not supported with set %s"
+			   " of settype %s\n",
+			   set->name, set->settype->typename);
+
 	while (offset < size) {
 		hash = (struct ip_set_hash_list *) (data + offset);
 		printf("%s -> %s\n", 
-			printip(hash->ip, options),
+			printip(set, hash->ip, options),
 			set_list[hash->binding]->name);
 		offset += sizeof(struct ip_set_hash_list);
 	}
@@ -1654,7 +1773,7 @@ static void print_bindings(void *data, size_t size, unsigned options,
 /* Help function to set_list() */
 static size_t print_set(void *data, unsigned options)
 {
-	struct ip_set_list *setlist = (struct ip_set_list *) data;
+	struct ip_set_list *setlist = data;
 	struct set *set = set_list[setlist->index];
 	struct settype *settype = set->settype;
 	size_t offset;
@@ -1688,7 +1807,9 @@ static size_t print_set(void *data, unsigned options)
 	/* Print bindings */
 	printf("Bindings:\n");
 	offset += setlist->members_size;
-	print_bindings(data + offset, setlist->bindings_size, options,
+	if (set->settype->bindip_tostring)
+		print_bindings(set,
+		       data + offset, setlist->bindings_size, options,
 		       settype->bindip_tostring);
 
 	printf("\n");		/* One newline between sets */
@@ -1700,20 +1821,24 @@ static int try_list_sets(const char name[IP_SET_MAXNAMELEN],
 			 unsigned options)
 {
 	void *data = NULL;
-	ip_set_id_t index;
-	size_t size, req_size;
+	ip_set_id_t idx;
+	socklen_t size, req_size;
 	int res = 0;
 
+	/* Default is numeric listing */
+	if (!(options & (OPT_RESOLVE|OPT_NUMERIC)))
+		options |= OPT_NUMERIC;
+
 	DP("%s", name);
 	/* Load set_list from kernel */
-	size = req_size = load_set_list(name, &index,
+	size = req_size = load_set_list(name, &idx,
 					IP_SET_OP_LIST_SIZE, CMD_LIST);
 
 	if (size) {
 		/* Get sets and print them */
 		data = ipset_malloc(size);
 		((struct ip_set_req_list *) data)->op = IP_SET_OP_LIST;
-		((struct ip_set_req_list *) data)->index = index;
+		((struct ip_set_req_list *) data)->index = idx;
 		res = kernel_getfrom_handleerrno(CMD_LIST, data, &size);
 		DP("get_lists getsockopt() res=%d errno=%d", res, errno);
 
@@ -1726,7 +1851,7 @@ static int try_list_sets(const char name[IP_SET_MAXNAMELEN],
 	while (size != req_size)
 		size += print_set(data + size, options);
 
-	ipset_free(&data);
+	ipset_free(data);
 	return res;
 }
 
@@ -1756,13 +1881,6 @@ static void list_sets(const char name[IP_SET_MAXNAMELEN], unsigned options)
  */
 static void set_help(const struct settype *settype)
 {
-#ifdef IPSET_DEBUG
-	char debughelp[] =
-	       "  --debug      -z   Enable debugging\n\n";
-#else
-	char debughelp[] = "\n";
-#endif
-
 	printf("%s v%s\n\n"
 	       "Usage: %s -N new-set settype [options]\n"
 	       "       %s -[XFLSH] [set] [options]\n"
@@ -1813,10 +1931,15 @@ static void set_help(const struct settype *settype)
 	       "                    Prints version information\n\n"
 	       "Options:\n"
 	       "  --sorted     -s   Numeric sort of the IPs in -L\n"
-	       "  --numeric    -n   Numeric output of addresses in a -L\n"
+	       "  --numeric    -n   Numeric output of addresses in a -L (default)\n"
+	       "  --resolve    -r   Try to resolve addresses in a -L\n"
 	       "  --quiet      -q   Suppress any output to stdout and stderr.\n"
 	       "  --binding    -b   Specifies the binding for -B\n");
-	printf(debughelp);
+#ifdef IPSET_DEBUG
+	printf("  --debug      -z   Enable debugging\n\n");
+#else
+	printf("\n");
+#endif
 
 	if (settype != NULL) {
 		printf("Type '%s' specific:\n", settype->typename);
@@ -1824,7 +1947,7 @@ static void set_help(const struct settype *settype)
 	}
 }
 
-static int find_cmd(const char option)
+static int find_cmd(int option)
 {
 	int i;
 	
@@ -1835,7 +1958,7 @@ static int find_cmd(const char option)
 	return CMD_NONE;
 }
 
-static int parse_adt_cmdline(unsigned command,
+static int parse_adt_cmdline(int command,
 			     const char *name,
 			     char *adt,
 			     struct set **set,
@@ -1885,7 +2008,7 @@ static int parse_adt_cmdline(unsigned command,
 int parse_commandline(int argc, char *argv[])
 {
 	int res = 0;
-	unsigned command = CMD_NONE;
+	int command = CMD_NONE;
 	unsigned options = 0;
 	int c;
 	
@@ -2039,6 +2162,7 @@ int parse_commandline(int argc, char *argv[])
 						   "Unknown arg `%s'",
 						   argv[optind - 1]);
 
+				res = 0;
 				break;
 			}
 
@@ -2048,6 +2172,11 @@ int parse_commandline(int argc, char *argv[])
 			add_option(&options, OPT_NUMERIC);
 			break;
 
+		case 'r':
+			if (!(options & OPT_NUMERIC))
+				add_option(&options, OPT_RESOLVE);
+			break;
+
 		case 's':
 			add_option(&options, OPT_SORTED);
 			break;
@@ -2071,7 +2200,7 @@ int parse_commandline(int argc, char *argv[])
 
 		case 1:	/* non option */
 			printf("Bad argument `%s'\n", optarg);
-			exit_tryhelp(2);
+			exit_tryhelp(PARAMETER_PROBLEM);
 			break;	/*always good */
 
 		default:{
@@ -2095,7 +2224,8 @@ int parse_commandline(int argc, char *argv[])
 					exit_error(PARAMETER_PROBLEM,
 						   "Unknown arg `%s'",
 						   argv[optind - 1]);
-
+				
+				res = 0;
 			}
 
 			DP("next arg");
@@ -2173,6 +2303,8 @@ int parse_commandline(int argc, char *argv[])
 		break;
 
 	case CMD_BIND:
+		fprintf(stderr, "Warning: binding will be removed from the next release.\n"
+			        "Please replace bindigs with sets of ipportmap and ipportiphash types\n");
 		if (restore)
 			set_restore_bind(set, adt, binding);
 		else
